/*
 * Copyright 2005-2009 NVIDIA Corporation.  All rights reserved.
 */

//-*-c++-*-
#ifdef USE_PCH
#include "opt_pch.h"
#endif // USE_PCH
#pragma hdrstop

#include "opt_ptrclass.h"
#include "wn_dfsolve.h"

// ====================================================================
//
// Description:
//
// This module implements pointer classification analysis to determine,
// using flow analysis, if a pointer points to global, local, constant,
// shared or parameter memory. It deals with the problem by relying on a
// forward data flow analysis formulated as follows:
//
// - all address taken operations are initialized to the kind of memory
//   that their symbols are stored in
// - all other pointers are initialized to top.
//
// Forward propagation them determines the kind of memory pointed to by
// each pointer, the meet operation moves down in the lattice which is
// as follows:
//
//           Top
//        / | | \\
//       g  l s c p
//       \  | | //
//         Bottom
//
// Note that if a pointer is uninitialized it may remain at Top and if
// a pointer may point to more than one memory space it may reach
// Bottom, the compiler should emit warnings if either condition is 
// encountered.
// 
// ====================================================================


static BOOL Tracing(void)
{
    return Get_Trace(TP_WOPT3, PTRCLASS_FLAG); 
}
 
char GetPtrClassesName(enum PtrClasses a)
{ 
    return "TBGLCSP"[(int)a]; 
}

// ====================================================================
//
// Description: Merge two pointer address spaces
//
// ====================================================================
enum PtrClasses MergePtrClasses(enum PtrClasses a, enum PtrClasses b) 
{
    if (a == Top || a == b)
        return b;
    if (b == Top)
        return a;
    return Bottom;
} // MergePtrClasses

// ====================================================================
//
// Description: Merge state with incoming state. Return TRUE if 
// state changed.
// ====================================================================
BOOL Pstate::Merge(Pstate *in)
{
    BOOL changed = FALSE;
    PtrClasses pclass;

    locmap_t *in_map = in->locmap;
    locmap_t::iterator itr, itr2;

    // For each location L from "in" Pstate:
    //      If L is absent in local state, insert into local state
    //      Else, merge L's ptrclass with local state's ptrclass.
    for (itr = in_map->begin(); itr != in_map->end(); itr++) {
        itr2 = locmap->find(itr->first);
        // not present in local state, insert
        if (itr2 == locmap->end()) {
            changed = TRUE;
            (*locmap)[itr->first] = itr->second;
        } else {
            // merge ptr classes. 
            pclass = MergePtrClasses(itr->second, itr2->second);
            if (pclass != itr2->second) {
                itr2->second = pclass;
                changed = TRUE;
            }
        }
    }
    // TRUE if state changed.
    return changed;
}// Pstate::Merge


// ====================================================================
//
// Description: Walk down a composite or scalar type. Insert each pointer
// variable found into the location map with the specified PtrClass.
// ====================================================================
void Pstate::WalkType(ST *root_st, TY_IDX ty, WN_OFFSET off, PtrClasses state)
{
    loc_t loc;
    FLD_ITER fld_iter;

    // found scalar pointer. Insert it into the location map
    if (TY_kind(ty) == KIND_POINTER)
    {
        loc.Set(root_st, off);
        (*locmap)[loc] = state;
         if (Tracing())
            fprintf(TFile,"initgen %p %d %c", loc.sym, (int)loc.off, GetPtrClassesName(state));
    } else if (TY_kind(ty) == KIND_STRUCT) {
        // found structure, iterate over fields.
        fld_iter = Make_fld_iter(TY_fld(ty)); 
        do {
            FLD_HANDLE fld(fld_iter);
            // if pointer, insert into location map
            if (TY_kind(FLD_type(fld)) == KIND_POINTER) {
                loc.Set(root_st, off + FLD_ofst(fld));
                (*locmap)[loc] = state;
                if (Tracing())
                    fprintf(TFile,"initgen %p %d %c", loc.sym, (int)loc.off, GetPtrClassesName(state));
            } else if (TY_kind(FLD_type(fld)) == KIND_STRUCT && 
                       TY_fld(FLD_type(fld)) != FLD_HANDLE()) 
            {
                // walk children fields.
                WalkType(root_st, FLD_type(fld), off + FLD_ofst(fld), state);
            }
        } while (!FLD_last_field(fld_iter++));
    }
} // Pstate::WalkType


// ====================================================================
//
// Description: Given a function entry node, initialize the state of the 
// pointers in variables to ptrclasses, as specified by CUDA semantics.
// ====================================================================
void Pstate::Initialize(WN *pu)
{
    WN *wn;
    ST *st;
    int ii;
    FmtAssert(WN_operator(pu) == OPR_FUNC_ENTRY, ("expect function entry node"));
    
    // TODO: bug fix : handle unions conservatively
    // Walk over the parameters
    for(ii = 0; ii <= (WN_kid_count(pu) - 4); ii++) {
        wn = WN_kid(pu, ii);
        FmtAssert((WN_operator(wn) == OPR_IDNAME), ("Formal argument must be OPR_IDNAME")); 
        st = WN_st(wn);
        WalkType(st, ST_type(st), 0, Global);
    }

    // Walk over global symtab
    // Look at pointer variables in constant and global space. They come
    // from the host code. The only space they can point to is the global
    // space. Just force them to point to global space in the locmap.
    FOREACH_SYMBOL(GLOBAL_SYMTAB, st, ii) {
        // pointers in constant memory must point to global memory
        if (ST_in_constant_mem(st) || ST_in_global_mem(st)) {
            WalkType(st, ST_type(st), 0, Global);
        }
    }
} // Pstate::Initialize

// ====================================================================
//
// Description: Reset the memory space for all pointers that are in
//              global and local symbol tables to bottom.
// ====================================================================
void Pstate::ResetToBottom(void)
{
    ST *st;
    int ii;

    FOREACH_SYMBOL(GLOBAL_SYMTAB, st, ii) {
      WalkType(st, ST_type(st), 0, Bottom);
    }
    FOREACH_SYMBOL(CURRENT_SYMTAB, st, ii) {
      WalkType(st, ST_type(st), 0, Bottom);
    }
} // Pstate::ResetToBottom

/////////////////////// start of PTR_CLASS methods
PTR_CLASS::PTR_CLASS(void)
{
    MEM_POOL_Initialize(&_ptr_pool, "ptr_class pool", FALSE);
    _expr_map = WN_MAP32_Create(&_ptr_pool);
}; // PTR_CLASS::PTR_CLASS

PTR_CLASS::~PTR_CLASS(void) 
{
    WN_MAP_Delete(_expr_map);
    MEM_POOL_Delete(&_ptr_pool);
}; // PTR_CLASS::~PTR_CLASS

// ====================================================================
//
// Description: Update the ptrclasses state attached to a WHIRL node.
//              Return TRUE if new state is different than old state
// ====================================================================
BOOL PTR_CLASS::UpdateExprMap(WN *wn, PtrClasses pstate)
{
    BOOL changed = FALSE;
    if (WN_MAP32_Get(_expr_map, wn) != pstate)
        changed = TRUE;
    WN_MAP32_Set(_expr_map, wn, (INT32)pstate);
    return changed;
}; // PTR_CLASS::UpdateExprMap
    
// ====================================================================
//
// Description: Apply the transfer function to "inp" and update "state"
// 
// ====================================================================
void PTR_CLASS::TransferFunction(WN *inp, Pstate *state)
{
    locmap_t::iterator itr;
    locmap_t *locmap;
    ST *st;
    loc_t loc;
    OPERATOR opr;
    INT32 t1, t2;
    enum PtrClasses pstate;
    BOOL changed;

    changed = FALSE;

    locmap = state->locmap;

    // TODO: bug fix: handle calls conservatively.
    opr = WN_operator(inp);
    if (opr == OPR_LDA) {
        // TODO: bug fix: handle address taken of symbol that is
        // a pointer or can contain a pointer conservatively.
        st = WN_st(inp);
        if (ST_class(st) == CLASS_VAR) {
            switch (ST_memory_space(st)) {
            case MEMORY_GLOBAL:
                pstate = Global;
                break;
            case MEMORY_SHARED:
                pstate = Shared;
                break;
            case MEMORY_LOCAL:
                pstate = Local;
                break;
            case MEMORY_CONSTANT:
                pstate = Constant;
                break;
            case MEMORY_PARAM:
                pstate = Param;
                break;
            case MEMORY_UNKNOWN:
                FmtAssert(ST_sclass(st) == SCLASS_AUTO || 
                          // MEMORY_UNKNOWN for SCLASS_FORMAL indicates the 
                          // formal parameter is a parameter of a device routine
                          // which is the same as SCLASS_AUTO.
                          ST_sclass(st) == SCLASS_FORMAL, 
                          ("Must be a local variable"));
                pstate = Local;
                break;
            default:
                FmtAssert(0, ("Must have a known memory space"));
                pstate = Bottom;
                break;
            }
        } else if (ST_class(st) == CLASS_CONST) {
            pstate = Constant;
        } else if (ST_class(st) == CLASS_FUNC) {
            pstate = Bottom;
        } else {
            FmtAssert(0, ("Unknown ST_Class for LDA"));
            pstate = Global;
        }
        if (Tracing())
            fprintf(TFile, "LDA: gen %p %c", inp, GetPtrClassesName(pstate));
        changed |= UpdateExprMap(inp, pstate);
    } else if (opr == OPR_ADD || opr == OPR_SUB) {
            t1 = WN_MAP32_Get(_expr_map, WN_kid0(inp));
            t2 = WN_MAP32_Get(_expr_map, WN_kid1(inp));
            pstate = MergePtrClasses((PtrClasses)t1, (PtrClasses)t2);
            if (UpdateExprMap(inp, pstate)) {
                changed = TRUE;
                if (Tracing()) {
                    fprintf(TFile, "%s: gen %p %c", opr == OPR_ADD ? "ADD" : "SUB", inp,
                            GetPtrClassesName(pstate));
                }
            }
            
    } else if (opr == OPR_SELECT) {
            t1 = WN_MAP32_Get(_expr_map, WN_kid1(inp));
            t2 = WN_MAP32_Get(_expr_map, WN_kid2(inp));
            pstate = MergePtrClasses((PtrClasses)t1, (PtrClasses)t2);
            if (UpdateExprMap(inp, pstate)) {
                changed = TRUE;
                if (Tracing())
                    fprintf(TFile, "SELECT: gen %p %c", inp, GetPtrClassesName(pstate));
            }
    } else if (opr == OPR_ASM_STMT) {
            pstate = Bottom;
            if (UpdateExprMap(inp, pstate)) {
                changed = TRUE;
                if (Tracing())
                    fprintf(TFile, "ASM_STMT: gen %p %c", inp, GetPtrClassesName(pstate));
            }
    } else if (opr == OPR_LDID) {
            loc.Set(WN_st(inp), WN_offset(inp));
            itr = locmap->find(loc);
            if (itr != locmap->end()) {
                changed |= UpdateExprMap(inp, itr->second);
                 if (Tracing())
                     fprintf(TFile, "LDID: gen %p %c", inp, GetPtrClassesName(itr->second));
            }
    } else if (opr == OPR_STID) {
            loc.Set(WN_st(inp), WN_offset(inp));
            itr = locmap->find(loc);
            pstate = (PtrClasses )WN_MAP32_Get(_expr_map, WN_kid0(inp));
            if (itr != locmap->end()) {
                changed |= (pstate != itr->second);
                if (Tracing()) {
                    fprintf(TFile,"STID: kill %p %d %c", WN_st(inp), (int)WN_offset(inp), 
                            GetPtrClassesName(itr->second));
                    fprintf(TFile,"STID: gen %p %d %c", WN_st(inp), (int)WN_offset(inp), 
                            GetPtrClassesName(pstate));
                }
                itr->second = pstate;
            } else if (pstate != Top) {// default state of whirl node is Top (0).
                // create on demand.
                (*locmap)[loc] = pstate;
                changed = TRUE;
                if (Tracing()) {
                    fprintf(TFile,"STID: gen %p %d %c", WN_st(inp), (int)WN_offset(inp),
                            GetPtrClassesName(pstate));
                }
            }
    } else if (opr == OPR_CALL) {
          // Conservatively assuming that a function call can assign to any pointer.
          // Therefore reset memory spaces of all pointers to bottom.
          state->ResetToBottom();
    }
}; // PTR_CLASS::TransferFunction

// ====================================================================
//
// Description: Do the ptr space analysis on the given function.
// 
// ====================================================================
void PTR_CLASS::Compute(WN *pu) 
{
    Pstate *initState = new Pstate();
    initState->Initialize(pu);
    DFSolver<PTR_CLASS, Pstate> solver(this);
    solver.Solve(TRUE, pu, initState);
    delete initState;
}; // PTR_CLASS::Compute

// ====================================================================
//
// Description: For a given ILOAD/ISTORE, return its PtrClass
// 
// ====================================================================
enum PtrClasses PTR_CLASS::Query(WN *inp)
{
    WN *wn;
    PtrClasses pstate;
    OPERATOR opr = WN_operator(inp);
            
    FmtAssert(opr == OPR_ILOAD || opr == OPR_ISTORE, ("Unexpected whirl node"));
    wn = (opr == OPR_ILOAD) ? WN_kid0(inp) : WN_kid1(inp);
    pstate = (PtrClasses)WN_MAP32_Get(_expr_map, wn);
    if (pstate == Top) // error condition? Top indicates uniniatialized memory
        pstate = Bottom;
    return pstate;
}; // PTR_CLASS::Query
